# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Signup @ https://www.sparkpost.com
#
# Ensure you've added a Senders Domain and have generated yourself an
# API Key at:
#   https://app.sparkpost.com/dashboard

# Note: For SMTP Access, your API key must have at least been granted the
#   'Send via SMTP' privileges.

# From here you can click on the domain you're interested in. You can acquire
# the API Key from here which will look something like:
#    1e1d479fcf1a87527e9411e083c700689fa1acdc
#
# Knowing this, you can buid your sparkpost url as follows:
#  sparkpost://{user}@{domain}/{apikey}
#  sparkpost://{user}@{domain}/{apikey}/{email}
#
# You can email as many addresses as you want as:
#  sparkpost://{user}@{domain}/{apikey}/{email1}/{email2}/{emailN}
#
#  The {user}@{domain} effectively assembles the 'from' email address
#  the email will be transmitted from.  If no email address is specified
#  then it will also become the 'to' address as well.
#
#  The {domain} must cross reference a domain you've set up with Spark Post
#
# API Documentation: https://developers.sparkpost.com/api/
# Specifically: https://developers.sparkpost.com/api/transmissions/
import contextlib
from email.utils import formataddr
from json import dumps, loads

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_bool, parse_emails, validate_regex
from .base import NotifyBase

# Provide some known codes SparkPost uses and what they translate to:
# Based on https://www.sparkpost.com/docs/tech-resources/extended-error-codes/
SPARKPOST_HTTP_ERROR_MAP = {
    400: "A bad request was made to the server",
    401: "Invalid User ID and/or Unauthorized User",
    403: "Permission Denied; the provided API Key was not valid",
    404: "There is a problem with the server query URI.",
    405: "Invalid HTTP method",
    420: "Sending limit reached.",
    422: "Invalid data/format/type/length",
    429: "To many requests per sec; rate limit",
}


class SparkPostRegion:
    """Regions."""

    US = "us"
    EU = "eu"


# SparkPost APIs
SPARKPOST_API_LOOKUP = {
    SparkPostRegion.US: "https://api.sparkpost.com/api/v1",
    SparkPostRegion.EU: "https://api.eu.sparkpost.com/api/v1",
}

# A List of our regions we can use for verification
SPARKPOST_REGIONS = (
    SparkPostRegion.US,
    SparkPostRegion.EU,
)


class NotifySparkPost(NotifyBase):
    """A wrapper for SparkPost Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "SparkPost"

    # The services URL
    service_url = "https://sparkpost.com/"

    # Support attachments
    attachment_support = True

    # All notification requests are secure
    secure_protocol = "sparkpost"

    # SparkPost advertises they allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.20

    # Words straight from their website:
    #    https://developers.sparkpost.com/api/#header-rate-limiting
    # These limits are dynamic, but as a general rule, wait 1 to 5 seconds
    # after receiving a 429 response before requesting again.

    # As a simple work around, this is what we will do... Wait X seconds
    # (defined below) before trying again when we get a 429 error
    sparkpost_retry_wait_sec = 5

    # The maximum number of times we'll retry to send our message when we've
    # reached a throttling situatin before giving up
    sparkpost_retry_attempts = 3

    # The maximum amount of emails that can reside within a single
    # batch transfer based on:
    #  https://www.sparkpost.com/docs/tech-resources/\
    #       smtp-rest-api-performance/#sending-via-the-transmission-rest-api
    default_batch_size = 2000

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_sparkpost"

    # Default Notify Format
    notify_format = NotifyFormat.HTML

    # Define object templates
    templates = (
        "{schema}://{user}@{host}:{apikey}/",
        "{schema}://{user}@{host}:{apikey}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "host": {
                "name": _("Domain"),
                "type": "string",
                "required": True,
            },
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "targets": {
                "name": _("Target Emails"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "name": {
                "name": _("From Name"),
                "type": "string",
                "map_to": "from_name",
            },
            "region": {
                "name": _("Region Name"),
                "type": "choice:string",
                "values": SPARKPOST_REGIONS,
                "default": SparkPostRegion.US,
                "map_to": "region_name",
            },
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("Email Header"),
            "prefix": "+",
        },
        "tokens": {
            "name": _("Template Tokens"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        apikey,
        targets,
        cc=None,
        bcc=None,
        from_name=None,
        region_name=None,
        headers=None,
        tokens=None,
        batch=None,
        **kwargs,
    ):
        """Initialize SparkPost Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid SparkPost API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate our username
        if not self.user:
            msg = "No SparkPost username was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire Email 'To'
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # For tracking our email -> name lookups
        self.names = {}

        # Store our region
        try:
            self.region_name = (
                self.template_args["region"]["default"]
                if region_name is None
                else region_name.lower()
            )

            if self.region_name not in SPARKPOST_REGIONS:
                # allow the outer except to handle this common response
                raise IndexError()

        except (AttributeError, IndexError, TypeError):
            # Invalid region specified
            msg = f"The SparkPost region specified ({region_name}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # Get our From username (if specified)
        self.from_name = from_name

        # Get our from email address
        self.from_addr = f"{self.user}@{self.host}"

        if not is_email(self.from_addr):
            # Parse Source domain based on from_addr
            msg = f"Invalid ~From~ email format: {self.from_addr}"
            self.logger.warning(msg)
            raise TypeError(msg)

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        self.tokens = {}
        if tokens:
            # Store our template tokens
            self.tokens.update(tokens)

        # Prepare Batch Mode Flag
        self.batch = (
            self.template_args["batch"]["default"] if batch is None else batch
        )

        if targets:
            # Validate recipients (to:) and drop bad ones:
            for recipient in parse_emails(targets):
                result = is_email(recipient)
                if result:
                    self.targets.append((
                        result["name"] if result["name"] else False,
                        result["full_email"],
                    ))
                    continue

                self.logger.warning(
                    f"Dropped invalid To email ({recipient}) specified.",
                )

        else:
            # If our target email list is empty we want to add ourselves to it
            self.targets.append(
                (self.from_name if self.from_name else False, self.from_addr)
            )

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            email = is_email(recipient)
            if email:
                self.cc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                f"Dropped invalid Carbon Copy email ({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):
            email = is_email(recipient)
            if email:
                self.bcc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                f"({recipient}) specified.",
            )

    def __post(self, payload, retry):
        """Performs the actual post and returns the response."""
        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": self.apikey,
        }

        # Prepare our URL as it's based on our hostname
        url = f"{SPARKPOST_API_LOOKUP[self.region_name]}/transmissions/"

        # Some Debug Logging
        self.logger.debug(
            "SparkPost POST URL:"
            f" {url} (cert_verify={self.verify_certificate})"
        )

        if "attachments" in payload["content"]:
            # Since we print our payload; attachments make it a bit too noisy
            # we just strip out the data block to accomodate it
            log_payload = {k: v for k, v in payload.items() if k != "content"}
            log_payload["content"] = {
                k: v
                for k, v in payload["content"].items()
                if k != "attachments"
            }
            log_payload["content"]["attachments"] = [
                {k: v for k, v in x.items() if k != "data"}
                for x in payload["content"]["attachments"]
            ]
        else:
            # No tidying is needed
            log_payload = payload

        self.logger.debug(f"SparkPost Payload: {log_payload}")

        wait = None

        # For logging output of success and errors; we get a head count
        # of our outbound details:
        verbose_dest = (
            ", ".join([x["address"]["email"] for x in payload["recipients"]])
            if len(payload["recipients"]) <= 3
            else "{} recipients".format(len(payload["recipients"]))
        )

        # Initialize our response object
        json_response = {}

        # Set ourselves a status code
        status_code = -1

        while 1:  # pragma: no branch

            # Always call throttle before any remote server i/o is made
            self.throttle(wait=wait)
            try:
                r = requests.post(
                    url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # A Good response (200) looks like this:
                #     "results": {
                #       "total_rejected_recipients": 0,
                #       "total_accepted_recipients": 1,
                #       "id": "11668787484950529"
                #        }
                #     }
                #
                # A Bad response looks like this:
                # {
                #   "errors": [
                #     {
                #       "description":
                #            "Unconfigured or unverified sending domain.",
                #       "code": "7001",
                #       "message": "Invalid domain"
                #     }
                #   ]
                # }
                #
                with contextlib.suppress(
                        AttributeError, TypeError, ValueError):
                    # Load our JSON Object if we can
                    # ValueError = r.content is Unparsable
                    # TypeError = r.content is None
                    # AttributeError = r is None
                    json_response = loads(r.content)

                status_code = r.status_code

                payload["recipients"] = []
                if status_code == requests.codes.ok:
                    self.logger.info(
                        f"Sent SparkPost notification to {verbose_dest}."
                    )
                    return status_code, json_response

                # We had a problem if we get here
                status_str = NotifyBase.http_response_code_lookup(
                    status_code, SPARKPOST_API_LOOKUP
                )

                self.logger.warning(
                    "Failed to send SparkPost notification to {}: "
                    "{}{}error={}.".format(
                        verbose_dest,
                        status_str,
                        ", " if status_str else "",
                        status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                if status_code == requests.codes.too_many_requests and retry:
                    retry = retry - 1
                    if retry > 0:
                        wait = self.sparkpost_retry_wait_sec
                        continue

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending SparkPost "
                    "notification"
                )
                self.logger.debug(f"Socket Exception: {e!s}")

            # Anything else and we're done
            return status_code, json_response

        # Our code will never reach here (outside of infinite while loop above)

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform SparkPost Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no SparkPost Email recipients to notify"
            )
            return False

        # Initialize our has_error flag
        has_error = False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        reply_to = formataddr(
            (self.from_name if self.from_name else False, self.from_addr),
            charset="utf-8",
        )

        payload = {
            "options": {
                # When set to True, an image is included with the email which
                # is used to detect if the user looked at the image or not.
                "open_tracking": False,
                # Track if links were clicked that were found within email
                "click_tracking": False,
            },
            "content": {
                "from": {
                    "name": (
                        self.from_name if self.from_name else self.app_desc
                    ),
                    "email": self.from_addr,
                },
                # SparkPost does not allow empty subject lines or lines that
                # only contain whitespace; Since Apprise allows an empty title
                # parameter we swap empty title entries with the period
                "subject": title if title.strip() else ".",
                "reply_to": reply_to,
            },
        }

        if self.notify_format == NotifyFormat.HTML:
            payload["content"]["html"] = body

        else:
            payload["content"]["text"] = body

        if attach and self.attachment_support:
            # Prepare ourselves an attachment object
            payload["content"]["attachments"] = []

            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SparkPost attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    # Prepare API Upload Payload
                    payload["content"]["attachments"].append({
                        "name": (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        ),
                        "type": attachment.mimetype,
                        "data": attachment.base64(),
                    })

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SparkPost attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending SparkPost attachment"
                    f" {attachment.url(privacy=True)}"
                )

        # Take a copy of our token dictionary
        tokens = self.tokens.copy()

        # Apply some defaults template values
        tokens["app_body"] = body
        tokens["app_title"] = title
        tokens["app_type"] = notify_type.value
        tokens["app_id"] = self.app_id
        tokens["app_desc"] = self.app_desc
        tokens["app_color"] = self.color(notify_type)
        tokens["app_url"] = self.app_url

        # Store our tokens if they're identified
        payload["substitution_data"] = self.tokens

        # Create a copy of the targets list
        emails = list(self.targets)

        for index in range(0, len(emails), batch_size):
            # Generate our email listing
            payload["recipients"] = []

            # Initialize our cc list
            cc = self.cc - self.bcc

            # Initialize our bcc list
            bcc = set(self.bcc)

            # Initialize our headers
            headers = self.headers.copy()

            for addr in self.targets[index : index + batch_size]:
                entry = {
                    "address": {
                        "email": addr[1],
                    }
                }

                # Strip target out of cc list if in To
                cc = cc - {addr[1]}

                # Strip target out of bcc list if in To
                bcc = bcc - {addr[1]}

                if addr[0]:
                    entry["address"]["name"] = addr[0]

                # Add our recipient to our list
                payload["recipients"].append(entry)

            if cc:
                # Handle our cc List
                for addr in cc:
                    entry = {
                        "address": {
                            "email": addr,
                            "header_to":
                            # Take the first email in the To
                            self.targets[index : index + batch_size][0][1],
                        },
                    }

                    if self.names.get(addr):
                        entry["address"]["name"] = self.names[addr]

                    # Add our recipient to our list
                    payload["recipients"].append(entry)

                headers["CC"] = ",".join(cc)

            # Handle our bcc
            for addr in bcc:
                # Add our recipient to our list
                payload["recipients"].append({
                    "address": {
                        "email": addr,
                        "header_to":
                        # Take the first email in the To
                        self.targets[index : index + batch_size][0][1],
                    },
                })

            if headers:
                payload["content"]["headers"] = headers

            # Send our message
            status_code, _response = self.__post(
                payload, self.sparkpost_retry_attempts
            )

            # Failed
            if status_code != requests.codes.ok:
                has_error = True

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user, self.apikey, self.host)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "region": self.region_name,
            "batch": "yes" if self.batch else "no",
        }

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Append our template tokens into our parameters
        params.update({f":{k}": v for k, v in self.tokens.items()})

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        if self.from_name is not None:
            # from_name specified; pass it back on the url
            params["name"] = self.from_name

        if self.cc:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                "{}{}".format(
                    "" if not e not in self.names else f"{self.names[e]}:",
                    e,
                )
                for e in self.cc
            ])

        if self.bcc:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(self.bcc)

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = not (
            len(self.targets) == 1 and self.targets[0][1] == self.from_addr
        )

        return "{schema}://{user}@{host}/{apikey}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            host=self.host,
            user=NotifySparkPost.quote(self.user, safe=""),
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets=(
                ""
                if not has_targets
                else "/".join([
                    NotifySparkPost.quote(
                        "{}{}".format("" if not e[0] else f"{e[0]}:", e[1]),
                        safe="",
                    )
                    for e in self.targets
                ])
            ),
            params=NotifySparkPost.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifySparkPost.split_path(results["fullpath"])

        # Our very first entry is reserved for our api key
        try:
            results["apikey"] = results["targets"].pop(0)

        except IndexError:
            # We're done - no API Key found
            results["apikey"] = None

        if "name" in results["qsd"] and len(results["qsd"]["name"]):
            # Extract from name to associate with from address
            results["from_name"] = NotifySparkPost.unquote(
                results["qsd"]["name"]
            )

        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            # Extract region
            results["region_name"] = NotifySparkPost.unquote(
                results["qsd"]["region"]
            )

        # Handle 'to' email address
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(results["qsd"]["to"])

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = results["qsd"]["cc"]

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = results["qsd"]["bcc"]

        # Add our Meta Headers that the user can provide with their outbound
        # emails
        results["headers"] = {
            NotifyBase.unquote(x): NotifyBase.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Add our template tokens (if defined)
        results["tokens"] = {
            NotifyBase.unquote(x): NotifyBase.unquote(y)
            for x, y in results["qsd:"].items()
        }

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifySparkPost.template_args["batch"]["default"]
            )
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Syntax:
#  schan://{key}/


class NotifyPushDeer(NotifyBase):
    """A wrapper for PushDeer Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "PushDeer"

    # The services URL
    service_url = "https://www.pushdeer.com/"

    # Insecure Protocol Access
    protocol = "pushdeer"

    # Secure Protocol
    secure_protocol = "pushdeers"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_PushDeer"

    # Default hostname
    default_hostname = "api2.pushdeer.com"

    # PushDeer API
    notify_url = "{schema}://{host}:{port}/message/push?pushkey={pushKey}"

    # Define object templates
    templates = (
        "{schema}://{pushkey}",
        "{schema}://{host}/{pushkey}",
        "{schema}://{host}:{port}/{pushkey}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "pushkey": {
                "name": _("Pushkey"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
        },
    )

    def __init__(self, pushkey, **kwargs):
        """Initialize PushDeer Object."""
        super().__init__(**kwargs)

        # PushKey (associated with project)
        self.push_key = validate_regex(
            pushkey, *self.template_tokens["pushkey"]["regex"]
        )
        if not self.push_key:
            msg = f"An invalid PushDeer API Pushkey ({pushkey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform PushDeer Notification."""

        # Prepare our persistent_notification.create payload
        payload = {
            "text": title if title else body,
            "type": "text",
            "desp": body if title else "",
        }

        # Set our schema
        schema = "https" if self.secure else "http"

        # Set host
        host = self.default_hostname
        if self.host:
            host = self.host

        # Set port
        port = 443 if self.secure else 80
        if self.port:
            port = self.port

        # Our Notification URL
        notify_url = self.notify_url.format(
            schema=schema, host=host, port=port, pushKey=self.push_key
        )

        # Some Debug Logging
        self.logger.debug(
            "PushDeer URL:"
            f" {notify_url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"PushDeer Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=payload,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyPushDeer.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send PushDeer notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return False

            else:
                self.logger.info("Sent PushDeer notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occured sending PushDeer notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.push_key,
            self.host,
            self.port,
        )

    def url(self, privacy=False):
        """Returns the URL built dynamically based on specified arguments."""

        if self.host:
            url = "{schema}://{host}{port}/{pushkey}"
        else:
            url = "{schema}://{pushkey}"

        return url.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            host=self.host,
            port="" if not self.port else f":{self.port}",
            pushkey=self.pprint(self.push_key, privacy, safe=""),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        substantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't parse the URL
            return results

        fullpaths = NotifyPushDeer.split_path(results["fullpath"])

        if len(fullpaths) == 0:
            results["pushkey"] = results["host"]
            results["host"] = None
        else:
            results["pushkey"] = fullpaths.pop()

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# You will need an API Key for this plugin to work.
# From the Settings -> API Keys you can click "Create API Key" if you don't
# have one already. The key must have at least the "Mail Send" permission
# to work.
#
# The schema to use the plugin looks like this:
#    {schema}://{apikey}:{from_addr}
#
# Your {from_addr} must be comprissed of your Resend Authenticated
# Domain.

# Simple API Reference:
#  - https://resend.com/onboarding

from email.utils import formataddr
from json import dumps

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_emails, validate_regex
from .base import NotifyBase

RESEND_HTTP_ERROR_MAP = {
    200: "Successful request.",
    400: "Check that the parameters were correct.",
    401: "The API key used was missing.",
    403: "The API key used was invalid.",
    404: "The resource was not found.",
    429: "The rate limit was exceeded.",
}


class NotifyResend(NotifyBase):
    """A wrapper for Notify Resend Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Resend"

    # The services URL
    service_url = "https://resend.com"

    # The default secure protocol
    secure_protocol = "resend"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_resend"

    # Default to markdown
    notify_format = NotifyFormat.HTML

    # The default Email API URL to use
    notify_url = "https://api.resend.com/emails"

    # Support attachments
    attachment_support = True

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # The default subject to use if one isn't specified.
    default_empty_subject = "<no subject>"

    # Define object templates
    templates = (
        "{schema}://{apikey}:{from_addr}",
        "{schema}://{apikey}:{from_addr}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9._-]+$", "i"),
            },
            "from_addr": {
                "name": _("Source Email"),
                "type": "string",
                "required": True,
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
            "reply": {
                "name": _("Reply To"),
                "type": "list:string",
                "map_to": "reply_to",
            },
            "from": {
                "map_to": "from_addr",
            },
            "name": {
                "name": _("From Name"),
                "map_to": "from_addr",
            },
            "apikey": {
                "map_to": "apikey",
            },
        },
    )

    def __init__(
        self, apikey, from_addr, targets=None, cc=None, bcc=None,
        reply_to=None, **kwargs):
        """Initialize Notify Resend Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Resend API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire Targets (To Emails)
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # Acquire Reply To
        self.reply_to = set()

        # For tracking our email -> name lookups
        self.names = {}

        result = is_email(from_addr)
        if not result:
            # Invalid from
            msg = "Invalid ~From~ email specified: {}".format(from_addr)
            self.logger.warning(msg)
            raise TypeError(msg)

        # initialize our from address
        self.from_addr = (
            result["name"] if result["name"] is not None else False,
            result["full_email"],
        )

        # Update our Name if specified
        self.names[self.from_addr[1]] = (
            result["name"] if result["name"] else False
        )

        # Acquire our targets
        targets = parse_emails(targets)
        if targets:
            # Validate recipients (to:) and drop bad ones:
            for recipient in targets:

                result = is_email(recipient)
                if result:
                    self.targets.append(result["full_email"])
                    continue

                self.logger.warning(
                    f"Dropped invalid email ({recipient}) specified.",
                )
        else:
            # If our target email list is empty we want to add ourselves to it
            self.targets.append(self.from_addr[1])

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):

            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])

                # Index our name (if one exists)
                self.names[result["full_email"]] = (
                    result["name"] if result["name"] else False
                )
                continue

            self.logger.warning(
                f"Dropped invalid Carbon Copy email ({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):

            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                f"({recipient}) specified.",
            )

        # Validate recipients (reply-to:) and drop bad ones:
        for recipient in parse_emails(reply_to):
            result = is_email(recipient)
            if result:
                self.reply_to.add(result["full_email"])

                # Index our name (if one exists)
                self.names[result["full_email"]] = (
                    result["name"] if result["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Reply To email ({}) specified.".format(
                    recipient
                ),
            )

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.from_addr)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.cc:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for it's escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8",
                ).replace(",", "%2C")
                for e in self.cc
            ])

        if len(self.bcc) > 0:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(self.bcc)

        if self.reply_to:
            # Handle our Reply-To Addresses
            params["reply"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for its escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8",
                ).replace(",", "%2C")
                for e in self.reply_to
            ])

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = not (
            len(self.targets) == 1 and self.targets[0] == self.from_addr[1])

        if self.from_addr[0] and self.from_addr[0] != self.app_id:
            # A custom name was provided
            params["name"] = self.from_addr[0]

        return "{schema}://{apikey}:{from_addr}/{targets}?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            # never encode email since it plays a huge role in our hostname
            from_addr=self.from_addr[1],
            targets=(
                ""
                if not has_targets
                else "/".join(
                    [NotifyResend.quote(x, safe="@") for x in self.targets]
                )
            ),
            params=NotifyResend.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Resend Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.apikey}",
        }

        # error tracking (used for function return)
        has_error = False

        # Prepare our from_name
        self.from_addr[0] \
            if self.from_addr[0] is not False else self.app_id

        _payload = {
            "from": formataddr(self.from_addr, charset="utf-8"),
            # A subject is a requirement, so if none is specified we must
            # set a default with at least 1 character or Resend will deny
            # our request
            "subject": title if title else self.default_empty_subject,
            (
                "text" if self.notify_format == NotifyFormat.TEXT else "html"
            ): body,
        }

        if attach and self.attachment_support:
            attachments = []

            # Send our attachments
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Resend attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    attachments.append({
                        "content": attachment.base64(),
                        "filename": (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        ),
                        "type": "application/octet-stream",
                        "disposition": "attachment",
                    })

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Resend attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Resend attachment"
                    f" {attachment.url(privacy=True)}"
                )

            # Append our attachments to the payload
            _payload.update({
                "attachments": attachments,
            })

        targets = list(self.targets)
        while len(targets) > 0:
            target = targets.pop(0)

            # Create a copy of our template
            payload = _payload.copy()

            # unique cc/bcc list management
            cc = self.cc - self.bcc - {target}
            bcc = self.bcc - {target}

            # handle our reply to
            reply_to = self.reply_to - {target}

            # Format our cc addresses to support the Name field
            cc = [
                formataddr((self.names.get(addr, False), addr),
                           charset="utf-8")
                for addr in cc
            ]

            # Format our reply-to addresses to support the Name field
            reply_to = [
                formataddr((self.names.get(addr, False), addr),
                           charset="utf-8")
                for addr in reply_to
            ]

            # Set our target
            payload["to"] = target

            if cc:
                payload["cc"] = cc

            if len(bcc):
                payload["bcc"] = list(bcc)

            if reply_to:
                payload["reply_to"] = reply_to

            self.logger.debug(
                "Resend POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug("Resend Payload: %s", str(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.accepted,
                ):
                    # We had a problem
                    status_str = NotifyResend.http_response_code_lookup(
                        r.status_code, RESEND_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send Resend notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Resend notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Resend "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Our URL looks like this:
        #    {schema}://{apikey}:{from_addr}/{targets}
        #
        # which actually equates to:
        #    {schema}://{user}:{password}@{host}/{email1}/{email2}/etc..
        #                 ^       ^         ^
        #                 |       |         |
        #              apikey     -from addr-

        # Prepare our API Key
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = \
                NotifyResend.unquote(results["qsd"]["apikey"])

        else:
            results["apikey"] = NotifyResend.unquote(results["user"])

        # Our Targets
        results["targets"] = []

        # Attempt to detect 'from' email address
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["from_addr"] = NotifyResend.unquote(results["qsd"]["from"])

            if results.get("host"):
                results["targets"].append(
                    NotifyResend.unquote(results["host"]))

        else:
            # Prepare our From Email Address
            results["from_addr"] = "{}@{}".format(
                NotifyResend.unquote(
                    results["password"]
                    if results["password"] else results["user"]),
                NotifyResend.unquote(results["host"]),
            )

        if "name" in results["qsd"] and len(results["qsd"]["name"]):
            results["from_addr"] = formataddr((
                NotifyResend.unquote(results["qsd"]["name"]),
                results["from_addr"]), charset="utf-8")

        # Acquire our targets
        results["targets"].extend(NotifyResend.split_path(results["fullpath"]))

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyResend.parse_list(results["qsd"]["to"])

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = NotifyResend.parse_list(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = NotifyResend.parse_list(results["qsd"]["bcc"])

        # Handle Reply To Addresses
        if "reply" in results["qsd"] and len(results["qsd"]["reply"]):
            results["reply_to"] = results["qsd"]["reply"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Emby Docker configuration: https://hub.docker.com/r/emby/embyserver/
# Authentication: https://github.com/MediaBrowser/Emby/wiki/Authentication
# Notifications: https://github.com/MediaBrowser/Emby/wiki/Remote-control
import hashlib
from json import dumps, loads

import requests

from .. import __version__ as VERSION
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool
from .base import NotifyBase


class NotifyEmby(NotifyBase):
    """A wrapper for Emby Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Emby"

    # The services URL
    service_url = "https://emby.media/"

    # The default protocol
    protocol = "emby"

    # The default secure protocol
    secure_protocol = "embys"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_emby"

    # By default Emby requires you to provide it a device id
    # The following was just a random uuid4 generated one.  There
    # is no real reason to change this, but hey; that's what open
    # source is for right?
    emby_device_id = "48df9504-6843-49be-9f2d-a685e25a0bc8"

    # The Emby message timeout; basically it is how long should our message be
    # displayed for.  The value is in milli-seconds
    emby_message_timeout_ms = 60000

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{user}:{password}@{host}",
        "{schema}://{user}:{password}@{host}:{port}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
                "default": 8096,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "modal": {
                "name": _("Modal"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(self, modal=False, **kwargs):
        """Initialize Emby Object."""
        super().__init__(**kwargs)

        if self.secure:
            self.schema = "https"

        else:
            self.schema = "http"

        # Our access token does not get created until we first
        # authenticate with our Emby server. The same goes for the
        # user id below.
        self.access_token = None
        self.user_id = None

        # Whether or not our popup dialog is a timed notification
        # or a modal type box (requires an Okay acknowledgement)
        self.modal = modal

        if not self.port:
            # Assign default port if one isn't otherwise specified:
            self.port = self.template_tokens["port"]["default"]

        if not self.user:
            # User was not specified
            msg = "No Emby username was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    def login(self, **kwargs):
        """Creates our authentication token and prepares our header."""

        if self.is_authenticated:
            # Log out first before we log back in
            self.logout()

        # Prepare our login url
        url = f"{self.schema}://{self.host}"
        if self.port:
            url += f":{self.port}"

        url += "/Users/AuthenticateByName"

        # Initialize our payload
        payload = {"Username": self.user}

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Emby-Authorization": self.emby_auth_header,
        }

        if self.password:
            # Source: https://github.com/MediaBrowser/Emby/wiki/Authentication
            # We require the following during our authentication
            #    pw - password in plain text
            #    password - password in Sha1
            #    passwordMd5 - password in MD5
            payload["pw"] = self.password

            password_md5 = hashlib.md5()
            password_md5.update(self.password.encode("utf-8"))
            payload["passwordMd5"] = password_md5.hexdigest()

            password_sha1 = hashlib.sha1()
            password_sha1.update(self.password.encode("utf-8"))
            payload["password"] = password_sha1.hexdigest()

        else:
            # Backwards compatibility
            payload["password"] = ""
            payload["passwordMd5"] = ""

            # April 1st, 2018 and newer requirement:
            payload["pw"] = ""

        self.logger.debug(
            "Emby login() POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )

        try:
            r = requests.post(
                url,
                headers=headers,
                data=dumps(payload),
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyEmby.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to authenticate Emby user {} details: "
                    "{}{}error={}.".format(
                        self.user,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred authenticating a user with Emby "
                f"at {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        # Load our results
        try:
            results = loads(r.content)

        except (AttributeError, TypeError, ValueError):
            # ValueError = r.content is Unparsable
            # TypeError = r.content is None
            # AttributeError = r is None

            # This is a problem; abort
            return False

        # Acquire our Access Token
        self.access_token = results.get("AccessToken")

        # Acquire our UserId. It can be in one (or both) of the
        # following locations in the response:
        #   {
        #      'User': {
        #         ...
        #         'Id': 'the_user_id_can_be_here',
        #         ...
        #       },
        #      'Id': 'the_user_id_can_be_found_here_too',
        #   }
        #
        # The below just safely covers both grounds.
        self.user_id = results.get("Id")
        if not self.user_id and "User" in results:
            self.user_id = results["User"].get("Id")

        # No user was found matching the specified
        return self.is_authenticated

    def sessions(self, user_controlled=True):
        """Acquire our Session Identifiers and store them in a dictionary
        indexed by the session id itself."""
        # A single session might look like this:
        # {
        #    u'AdditionalUsers': [],
        #    u'ApplicationVersion': u'3.3.1.0',
        #    u'Client': u'Emby Mobile',
        #    u'DeviceId': u'00c901e90ae814c00f81c75ae06a1c8a4381f45b',
        #    u'DeviceName': u'Firefox',
        #    u'Id': u'e37151ea06d7eb636639fded5a80f223',
        #    u'LastActivityDate': u'2018-03-04T21:29:02.5590200Z',
        #    u'PlayState': {
        #       u'CanSeek': False,
        #       u'IsMuted': False,
        #       u'IsPaused': False,
        #       u'RepeatMode': u'RepeatNone',
        #    },
        #    u'PlayableMediaTypes': [u'Audio', u'Video'],
        #    u'RemoteEndPoint': u'172.17.0.1',
        #    u'ServerId': u'4470e977ea704a08b264628c24127d43',
        #    u'SupportedCommands': [
        #       u'MoveUp',
        #       u'MoveDown',
        #       u'MoveLeft',
        #       u'MoveRight',
        #       u'PageUp',
        #       u'PageDown',
        #       u'PreviousLetter',
        #       u'NextLetter',
        #       u'ToggleOsd',
        #       u'ToggleContextMenu',
        #       u'Select',
        #       u'Back',
        #       u'SendKey',
        #       u'SendString',
        #       u'GoHome',
        #       u'GoToSettings',
        #       u'VolumeUp',
        #       u'VolumeDown',
        #       u'Mute',
        #       u'Unmute',
        #       u'ToggleMute',
        #       u'SetVolume',
        #       u'SetAudioStreamIndex',
        #       u'SetSubtitleStreamIndex',
        #       u'DisplayContent',
        #       u'GoToSearch',
        #       u'DisplayMessage',
        #       u'SetRepeatMode',
        #       u'ChannelUp',
        #       u'ChannelDown',
        #       u'PlayMediaSource',
        #    ],
        #    u'SupportsRemoteControl': True,
        #    u'UserId': u'6f98d12cb10f48209ee282787daf7af6',
        #    u'UserName': u'l2g'
        #    }

        # Prepare a dict() object to control our sessions; the keys are
        # the sessions while the details associated with the session
        # are stored inside.
        sessions = {}

        if not self.is_authenticated and not self.login():
            # Authenticate if we aren't already
            return sessions

        # Prepare our login url
        url = f"{self.schema}://{self.host}"
        if self.port:
            url += f":{self.port}"

        url += "/Sessions"

        if user_controlled is True:
            # Only return sessions that can be managed by the current Emby
            # user.
            url += f"?ControllableByUserId={self.user_id}"

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Emby-Authorization": self.emby_auth_header,
            "X-MediaBrowser-Token": self.access_token,
        }

        self.logger.debug(
            "Emby session() GET URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )

        try:
            r = requests.get(
                url,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyEmby.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to acquire Emby session for user {}: "
                    "{}{}error={}.".format(
                        self.user,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return sessions

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred querying Emby "
                f"for session information at {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return sessions

        # Load our results
        try:
            results = loads(r.content)

        except (AttributeError, TypeError, ValueError):
            # ValueError = r.content is Unparsable
            # TypeError = r.content is None
            # AttributeError = r is None

            # We need to abort at this point
            return sessions

        for entry in results:
            session = entry.get("Id")
            if session:
                sessions[session] = entry

        return sessions

    def logout(self, **kwargs):
        """Logs out of an already-authenticated session."""
        if not self.is_authenticated:
            # We're not authenticated; there is nothing to do
            return True

        # Prepare our login url
        url = f"{self.schema}://{self.host}"
        if self.port:
            url += f":{self.port}"

        url += "/Sessions/Logout"

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Emby-Authorization": self.emby_auth_header,
            "X-MediaBrowser-Token": self.access_token,
        }

        self.logger.debug(
            "Emby logout() POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        try:
            r = requests.post(
                url,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code not in (
                # We're already logged out
                requests.codes.unauthorized,
                # The below show up if we were 'just' logged out
                requests.codes.ok,
                requests.codes.no_content,
            ):

                # We had a problem
                status_str = NotifyEmby.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to logoff Emby user {}: {}{}error={}.".format(
                        self.user,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred querying Emby "
                f"to logoff user {self.user} at {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        # We logged our successfully if we reached here

        # Reset our variables
        self.access_token = None
        self.user_id = None
        return True

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Emby Notification."""
        if not self.is_authenticated and not self.login():
            # Authenticate if we aren't already
            return False

        # Acquire our list of sessions
        sessions = self.sessions().keys()
        if not sessions:
            self.logger.warning("There were no Emby sessions to notify.")
            # We don't need to fail; there really is no one to notify
            return True

        url = f"{self.schema}://{self.host}"
        if self.port:
            url += f":{self.port}"

        # Append our remaining path
        url += "/Sessions/%s/Message"

        # Prepare Emby Object
        payload = {
            "Header": title,
            "Text": body,
        }

        if not self.modal:
            payload["TimeoutMs"] = self.emby_message_timeout_ms

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Emby-Authorization": self.emby_auth_header,
            "X-MediaBrowser-Token": self.access_token,
        }

        # Track whether or not we had a failure or not.
        has_error = False

        for session in sessions:
            # Update our session
            session_url = url % session

            self.logger.debug(
                "Emby POST URL:"
                f" {session_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Emby Payload: {payload!s}")

            # Always call throttle before the requests are made
            self.throttle()

            try:
                r = requests.post(
                    session_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.no_content,
                ):
                    # We had a problem
                    status_str = NotifyEmby.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Emby notification: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info("Sent Emby notification.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Emby "
                    f"notification to {self.host}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "modal": "yes" if self.modal else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyEmby.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        else:  # self.user is set
            auth = "{user}@".format(
                user=NotifyEmby.quote(self.user, safe=""),
            )

        return "{schema}://{auth}{hostname}{port}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            hostname=self.host,
            port=(
                ""
                if self.port is None
                or self.port == self.template_tokens["port"]["default"]
                else f":{self.port}"
            ),
            params=NotifyEmby.urlencode(params),
        )

    @property
    def is_authenticated(self):
        """Returns True if we're authenticated and False if not."""
        return bool(self.access_token and self.user_id)

    @property
    def emby_auth_header(self):
        """Generates the X-Emby-Authorization header response based on whether
        we're authenticated or not."""
        # Specific to Emby
        header_args = [
            ("MediaBrowser Client", self.app_id),
            ("Device", self.app_id),
            ("DeviceId", self.emby_device_id),
            ("Version", str(VERSION)),
        ]

        if self.user_id:
            # Append UserId variable if we're authenticated
            header_args.append(("UserId", self.user))

        return ", ".join([f'{k}="{v}"' for k, v in header_args])

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early
            return results

        # Modal type popup (default False)
        results["modal"] = parse_bool(results["qsd"].get("modal", False))

        return results

    def __del__(self):
        """Destructor."""
        self.logout()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# PAHO MQTT Documentation:
#  https://www.eclipse.org/paho/index.php?page=clients/python/docs/index.php
#
# Looking at the PAHO MQTT Source can help shed light on what's going on too
# as their inline documentation is pretty good!
#   https://github.com/eclipse/paho.mqtt.python\
#           /blob/master/src/paho/mqtt/client.py
from datetime import datetime
from os.path import isfile
import re
import ssl
from time import sleep

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list
from .base import NotifyBase

# Default our global support flag
NOTIFY_MQTT_SUPPORT_ENABLED = False

try:
    # 3rd party modules
    import paho.mqtt.client as mqtt

    # We're good to go!
    NOTIFY_MQTT_SUPPORT_ENABLED = True

    MQTT_PROTOCOL_MAP = {
        # v3.1.1
        "311": mqtt.MQTTv311,
        # v3.1
        "31": mqtt.MQTTv31,
        # v5.0
        "5": mqtt.MQTTv5,
        # v5.0 (alias)
        "50": mqtt.MQTTv5,
    }

except ImportError:
    # No problem; we just simply can't support this plugin because we're
    # either using Linux, or simply do not have pywin32 installed.
    MQTT_PROTOCOL_MAP = {}

# A lookup map for relaying version to user
HUMAN_MQTT_PROTOCOL_MAP = {
    "v3.1.1": "311",
    "v3.1": "31",
    "v5.0": "5",
}


class NotifyMQTT(NotifyBase):
    """A wrapper for MQTT Notifications."""

    # Set our global enabled flag
    enabled = NOTIFY_MQTT_SUPPORT_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "paho-mqtt != 2.0.*"
    }

    # The default descriptive name associated with the Notification
    service_name = "MQTT Notification"

    # The default protocol
    protocol = "mqtt"

    # Secure protocol
    secure_protocol = "mqtts"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_mqtt"

    # MQTT does not have a title
    title_maxlen = 0

    # The maximum length a body can be set to
    body_maxlen = 268435455

    # Use a throttle; but it doesn't need to be so strict since most
    # MQTT server hostings can handle the small bursts of packets and are
    # locally hosted anyway
    request_rate_per_sec = 0.5

    # Port Defaults (unless otherwise specified)
    mqtt_insecure_port = 1883

    # The default secure port to use (if mqtts://)
    mqtt_secure_port = 8883

    # The default mqtt keepalive value
    mqtt_keepalive = 30

    # The default mqtt transport
    mqtt_transport = "tcp"

    # The number of seconds to wait for a publish to occur at before
    # checking to see if it's been sent yet.
    mqtt_block_time_sec = 0.2

    # Set the maximum number of messages with QoS>0 that can be part way
    # through their network flow at once.
    mqtt_inflight_messages = 200

    # Define object templates
    templates = (
        "{schema}://{user}@{host}/{topic}",
        "{schema}://{user}@{host}:{port}/{topic}",
        "{schema}://{user}:{password}@{host}/{topic}",
        "{schema}://{user}:{password}@{host}:{port}/{topic}",
    )

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "topic": {
                "name": _("Target Queue"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "qos": {
                "name": _("QOS"),
                "type": "int",
                "default": 0,
                "min": 0,
                "max": 2,
            },
            "version": {
                "name": _("Version"),
                "type": "choice:string",
                "values": HUMAN_MQTT_PROTOCOL_MAP,
                "default": "v3.1.1",
            },
            "client_id": {
                "name": _("Client ID"),
                "type": "string",
            },
            "session": {
                "name": _("Use Session"),
                "type": "bool",
                "default": False,
            },
            "retain": {
                "name": _("Retain Messages"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self,
        targets=None,
        version=None,
        qos=None,
        client_id=None,
        session=None,
        retain=None,
        **kwargs,
    ):
        """Initialize MQTT Object."""

        super().__init__(**kwargs)

        # Initialize topics
        self.topics = parse_list(targets)

        if version is None:
            self.version = self.template_args["version"]["default"]
        else:
            self.version = version

        # Save our client id if specified
        self.client_id = client_id

        # Maintain our session (associated with our user id if set)
        self.session = (
            self.template_args["session"]["default"]
            if session is None or not self.client_id
            else parse_bool(session)
        )

        # Our Retain Message Flag
        self.retain = (
            self.template_args["retain"]["default"]
            if retain is None
            else parse_bool(retain)
        )

        # Set up our Quality of Service (QoS)
        try:
            self.qos = (
                self.template_args["qos"]["default"]
                if qos is None
                else int(qos)
            )

            if (
                self.qos < self.template_args["qos"]["min"]
                or self.qos > self.template_args["qos"]["max"]
            ):
                # Let error get handle on exceptio higher up
                raise ValueError("")

        except (ValueError, TypeError):
            msg = f"An invalid MQTT QOS ({qos}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        if not self.port:
            # Assign port (if not otherwise set)
            self.port = (
                self.mqtt_secure_port
                if self.secure
                else self.mqtt_insecure_port
            )

        self.ca_certs = None
        if self.secure:
            # verify SSL key or abort
            # TODO: There is no error reporting or aborting here?
            #       It could be useful to inform the user _where_ Apprise
            #       tried to find the root CA certificates file.
            self.ca_certs = next(
                (
                    cert
                    for cert in self.CA_CERTIFICATE_FILE_LOCATIONS
                    if isfile(cert)
                ),
                None,
            )

        # Set up our MQTT Publisher
        try:
            # Get our protocol
            self.mqtt_protocol = MQTT_PROTOCOL_MAP[
                re.sub(r"[^0-9]+", "", self.version)
            ]

        except KeyError:
            msg = (
                f"An invalid MQTT Protocol version ({version}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # Our MQTT Client Object
        self.client = mqtt.Client(
            client_id=self.client_id,
            clean_session=not self.session,
            userdata=None,
            protocol=self.mqtt_protocol,
            transport=self.mqtt_transport,
        )

        # Our maximum number of in-flight messages
        self.client.max_inflight_messages_set(self.mqtt_inflight_messages)

        # Toggled to False once our connection has been established at least
        # once
        self.__initial_connect = True

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform MQTT Notification."""

        if len(self.topics) == 0:
            # There were no services to notify
            self.logger.warning("There were no MQTT topics to notify.")
            return False

        # For logging:
        url = f"{self.host}:{self.port}"

        try:
            if self.__initial_connect:
                # Our initial connection
                if self.user:
                    self.client.username_pw_set(
                        self.user, password=self.password
                    )

                if self.secure:
                    if self.ca_certs is None:
                        self.logger.error(
                            "MQTT secure communication can not be verified, "
                            "CA certificates file missing"
                        )
                        return False

                    self.client.tls_set(
                        ca_certs=self.ca_certs,
                        certfile=None,
                        keyfile=None,
                        cert_reqs=ssl.CERT_REQUIRED,
                        tls_version=ssl.PROTOCOL_TLS,
                        ciphers=None,
                    )

                    # Set our TLS Verify Flag
                    self.client.tls_insecure_set(not self.verify_certificate)

                # Establish our connection
                if (
                    self.client.connect(
                        self.host,
                        port=self.port,
                        keepalive=self.mqtt_keepalive,
                    )
                    != mqtt.MQTT_ERR_SUCCESS
                ):
                    self.logger.warning(
                        "An MQTT connection could not be established for"
                        f" {url}"
                    )
                    return False

                # Start our client loop
                self.client.loop_start()

                # Throttle our start otherwise the starting handshaking doesnt
                # work. I'm not sure if this is a bug or not, but with qos=0,
                # and without this sleep(), the messages randomly fails to be
                # delivered.
                sleep(0.01)

                # Toggle our flag since we never need to enter this area again
                self.__initial_connect = False

            # Create a copy of the subreddits list
            topics = list(self.topics)

            has_error = False
            while len(topics) > 0 and not has_error:
                # Retrieve our subreddit
                topic = topics.pop()

                # For logging:
                url = f"{self.host}:{self.port}/{topic}"

                # Always call throttle before any remote server i/o is made
                self.throttle()

                # handle a re-connection
                if (
                    not self.client.is_connected()
                    and self.client.reconnect() != mqtt.MQTT_ERR_SUCCESS
                ):
                    self.logger.warning(
                        f"An MQTT connection could not be sustained for {url}"
                    )
                    has_error = True
                    break

                # Some Debug Logging
                self.logger.debug(
                    "MQTT POST URL:"
                    f" {url} (cert_verify={self.verify_certificate})"
                )
                self.logger.debug(f"MQTT Payload: {body!s}")

                result = self.client.publish(
                    topic, payload=body, qos=self.qos, retain=self.retain
                )

                if result.rc != mqtt.MQTT_ERR_SUCCESS:
                    # Toggle our status
                    self.logger.warning(
                        f"An error (rc={result.rc}) occured when sending MQTT"
                        f" to {url}"
                    )
                    has_error = True
                    break

                elif not result.is_published():
                    self.logger.debug(
                        "Blocking until MQTT payload is published..."
                    )
                    reference = datetime.now()
                    while not has_error and not result.is_published():
                        # Throttle
                        sleep(self.mqtt_block_time_sec)

                        # Our own throttle so we can abort eventually....
                        elapsed = (datetime.now() - reference).total_seconds()
                        if elapsed >= self.socket_read_timeout:
                            self.logger.warning(
                                "The MQTT message could not be delivered"
                            )
                            has_error = True

                # if we reach here; we're at the bottom of our loop
                # we loop around and do the next topic now

        except ConnectionError as e:
            self.logger.warning(f"MQTT Connection Error received from {url}")
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except ssl.CertificateError as e:
            self.logger.warning(
                f"MQTT SSL Certificate Error received from {url}"
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except ValueError as e:
            # ValueError's are thrown from publish() call if there is a problem
            self.logger.warning(f"MQTT Publishing error received: from {url}")
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        if not has_error:
            # Verbal notice
            self.logger.info("Sent MQTT notification")

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            (
                self.port
                if self.port
                else (
                    self.mqtt_secure_port
                    if self.secure
                    else self.mqtt_insecure_port
                )
            ),
            self.fullpath.rstrip("/"),
            self.client_id,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "version": self.version,
            "qos": str(self.qos),
            "session": "yes" if self.session else "no",
            "retain": "yes" if self.retain else "no",
        }

        if self.client_id:
            # Our client id is set if specified
            params["client_id"] = self.client_id

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyMQTT.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyMQTT.quote(self.user, safe=""),
            )

        default_port = (
            self.mqtt_secure_port if self.secure else self.mqtt_insecure_port
        )

        return "{schema}://{auth}{hostname}{port}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            targets=",".join(
                [NotifyMQTT.quote(x, safe="/") for x in self.topics]
            ),
            params=NotifyMQTT.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.topics)

    @staticmethod
    def parse_url(url):
        """There are no parameters nessisary for this protocol; simply having
        windows:// is all you need.

        This function just makes sure that is in place.
        """

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        try:
            # Acquire topic(s)
            results["targets"] = parse_list(
                NotifyMQTT.unquote(results["fullpath"].lstrip("/"))
            )

        except AttributeError:
            # No 'fullpath' specified
            results["targets"] = []

        # The MQTT protocol version to use
        if "version" in results["qsd"] and len(results["qsd"]["version"]):
            results["version"] = NotifyMQTT.unquote(results["qsd"]["version"])

        # The MQTT Client ID
        if "client_id" in results["qsd"] and len(results["qsd"]["client_id"]):
            results["client_id"] = NotifyMQTT.unquote(
                results["qsd"]["client_id"]
            )

        if "session" in results["qsd"] and len(results["qsd"]["session"]):
            results["session"] = parse_bool(results["qsd"]["session"])

        # Message Retain Flag
        if "retain" in results["qsd"] and len(results["qsd"]["retain"]):
            results["retain"] = parse_bool(results["qsd"]["retain"])

        # The MQTT Quality of Service to use
        if "qos" in results["qsd"] and len(results["qsd"]["qos"]):
            results["qos"] = NotifyMQTT.unquote(results["qsd"]["qos"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].extend(
                NotifyMQTT.parse_list(results["qsd"]["to"])
            )

        # return results
        return results

    @property
    def CA_CERTIFICATE_FILE_LOCATIONS(self):
        """Return possible locations to root certificate authority (CA)
        bundles.

        Taken from https://golang.org/src/crypto/x509/root_linux.go
        TODO: Maybe refactor to a general utility function?
        """
        candidates = [
            # Debian/Ubuntu/Gentoo etc.
            "/etc/ssl/certs/ca-certificates.crt",
            # Fedora/RHEL 6
            "/etc/pki/tls/certs/ca-bundle.crt",
            # OpenSUSE
            "/etc/ssl/ca-bundle.pem",
            # OpenELEC
            "/etc/pki/tls/cacert.pem",
            # CentOS/RHEL 7
            "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",
            # macOS Homebrew; brew install ca-certificates
            "/usr/local/etc/ca-certificates/cert.pem",
        ]

        # Certifi provides Mozilla's carefully curated collection of Root
        # Certificates for validating the trustworthiness of SSL certificates
        # while verifying the identity of TLS hosts. It has been extracted from
        # the Requests project.
        try:
            import certifi

            candidates.append(certifi.where())
        except ImportError:  # pragma: no cover
            pass

        return candidates



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a Sinch account to which you can get your
# API_TOKEN and SERVICE_PLAN_ID right from your console/dashboard at:
#     https://dashboard.sinch.com/sms/overview
#
# You will also need to send the SMS From a phone number or account id name.

# This is identified as the source (or where the SMS message will originate
# from). Activated phone numbers can be found on your dashboard here:
#  - https://dashboard.sinch.com/numbers/your-numbers/numbers
#
import json

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class SinchRegion:
    """Defines the Sinch Server Regions."""

    USA = "us"
    EUROPE = "eu"


# Used for verification purposes
SINCH_REGIONS = (SinchRegion.USA, SinchRegion.EUROPE)


class NotifySinch(NotifyBase):
    """A wrapper for Sinch Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Sinch"

    # The services URL
    service_url = "https://sinch.com/"

    # All notification requests are secure
    secure_protocol = "sinch"

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.20

    # the number of seconds undelivered messages should linger for
    # in the Sinch queue
    validity_period = 14400

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_sinch"

    # Sinch uses the http protocol with JSON requests
    #   - the 'spi' gets substituted with the Service Provider ID
    #     provided as part of the Apprise URL.
    notify_url = "https://{region}.sms.api.sinch.com/xms/v1/{spi}/batches"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{service_plan_id}:{api_token}@{from_phone}",
        "{schema}://{service_plan_id}:{api_token}@{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "service_plan_id": {
                "name": _("Account SID"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-f0-9]+$", "i"),
            },
            "api_token": {
                "name": _("Auth Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-f0-9]+$", "i"),
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "required": True,
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "short_code": {
                "name": _("Target Short Code"),
                "type": "string",
                "regex": (r"^[0-9]{5,6}$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
            "spi": {
                "alias_of": "service_plan_id",
            },
            "region": {
                "name": _("Region"),
                "type": "string",
                "regex": (r"^[a-z]{2}$", "i"),
                "default": SinchRegion.USA,
            },
            "token": {
                "alias_of": "api_token",
            },
        },
    )

    def __init__(
        self,
        service_plan_id,
        api_token,
        source,
        targets=None,
        region=None,
        **kwargs,
    ):
        """Initialize Sinch Object."""
        super().__init__(**kwargs)

        # The Account SID associated with the account
        self.service_plan_id = validate_regex(
            service_plan_id, *self.template_tokens["service_plan_id"]["regex"]
        )
        if not self.service_plan_id:
            msg = (
                "An invalid Sinch Account SID "
                f"({service_plan_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # The Authentication Token associated with the account
        self.api_token = validate_regex(
            api_token, *self.template_tokens["api_token"]["regex"]
        )
        if not self.api_token:
            msg = (
                "An invalid Sinch Authentication Token "
                f"({api_token}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Setup our region
        self.region = (
            self.template_args["region"]["default"]
            if not isinstance(region, str)
            else region.lower()
        )
        if self.region and self.region not in SINCH_REGIONS:
            msg = f"The region specified ({region}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # The Source Phone # and/or short-code
        result = is_phone_no(source, min_len=5)
        if not result:
            msg = (
                "The Account (From) Phone # or Short-code specified "
                f"({source}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Tidy source
        self.source = result["full"]

        if len(self.source) < 11 or len(self.source) > 14:
            # A short code is a special 5 or 6 digit telephone number
            # that's shorter than a full phone number.
            if len(self.source) not in (5, 6):
                msg = (
                    "The Account (From) Phone # specified "
                    f"({source}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            # else... it as a short code so we're okay

        else:
            # We're dealing with a phone number; so we need to just
            # place a plus symbol at the end of it
            self.source = f"+{self.source}"

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Parse each phone number we found
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append("+{}".format(result["full"]))

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Sinch Notification."""

        if not self.targets and len(self.source) in (5, 6):
            # Generate a warning since we're a short-code.  We need
            # a number to message at minimum
            self.logger.warning("There are no valid Sinch targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Authorization": f"Bearer {self.api_token}",
            "Content-Type": "application/json",
        }

        # Prepare our payload
        payload = {
            "body": body,
            "from": self.source,
            # The To gets populated in the loop below
            "to": None,
        }

        # Prepare our Sinch URL (spi = Service Provider ID)
        url = self.notify_url.format(
            region=self.region, spi=self.service_plan_id
        )

        # Create a copy of the targets list
        targets = list(self.targets)

        if len(targets) == 0:
            # No sources specified, use our own phone no
            targets.append(self.source)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload["to"] = [target]

            # Some Debug Logging
            self.logger.debug(
                "Sinch POST URL:"
                f" {url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Sinch Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    url,
                    data=json.dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # The responsne might look like:
                # {
                #  "id": "CJloRJOe3MtDITqx",
                #  "to": ["15551112222"],
                #  "from": "15553334444",
                #  "canceled": false,
                #  "body": "This is a test message from your Sinch account",
                #  "type": "mt_text",
                #  "created_at": "2020-01-14T01:05:20.694Z",
                #  "modified_at": "2020-01-14T01:05:20.694Z",
                #  "delivery_report": "none",
                #  "expire_at": "2020-01-17T01:05:20.694Z",
                #  "flash_message": false
                # }
                if r.status_code not in (
                    requests.codes.created,
                    requests.codes.ok,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code
                    )

                    # set up our status code to use
                    status_code = r.status_code

                    try:
                        # Update our status response if we can
                        json_response = json.loads(r.content)
                        status_code = json_response.get("code", status_code)
                        status_str = json_response.get("message", status_str)

                    except (AttributeError, TypeError, ValueError):
                        # ValueError = r.content is Unparsable
                        # TypeError = r.content is None
                        # AttributeError = r is None

                        # We could not parse JSON response.
                        # We will just use the status we already have.
                        pass

                    self.logger.warning(
                        "Failed to send Sinch notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Sinch notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending Sinch:{target} "
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.service_plan_id,
            self.api_token,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "region": self.region,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{spi}:{token}@{source}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            spi=self.pprint(
                self.service_plan_id, privacy, mode=PrivacyMode.Tail, safe=""
            ),
            token=self.pprint(self.api_token, privacy, safe=""),
            source=NotifySinch.quote(self.source, safe=""),
            targets="/".join(
                [NotifySinch.quote(x, safe="") for x in self.targets]
            ),
            params=NotifySinch.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifySinch.split_path(results["fullpath"])

        # The hostname is our source number
        results["source"] = NotifySinch.unquote(results["host"])

        # Get our service_plan_ide and api_token from the user/pass config
        results["service_plan_id"] = NotifySinch.unquote(results["user"])
        results["api_token"] = NotifySinch.unquote(results["password"])

        # Auth Token
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Extract the account spi from an argument
            results["api_token"] = NotifySinch.unquote(results["qsd"]["token"])

        # Account SID
        if "spi" in results["qsd"] and len(results["qsd"]["spi"]):
            # Extract the account spi from an argument
            results["service_plan_id"] = NotifySinch.unquote(
                results["qsd"]["spi"]
            )

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifySinch.unquote(results["qsd"]["from"])

        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifySinch.unquote(results["qsd"]["source"])

        # Allow one to define a region
        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            results["region"] = NotifySinch.unquote(results["qsd"]["region"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySinch.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# New priorities are defined here:
# - https://firebase.google.com/docs/reference/fcm/rest/v1/\
#       projects.messages#NotificationPriority

# Legacy color payload example here:
# https://firebase.google.com/docs/reference/fcm/rest/v1/\
#       projects.messages#androidnotification
import re

from ...asset import AppriseAsset
from ...common import NotifyType
from ...utils.parse import parse_bool


class FCMColorManager:
    """A Simple object to accept either a boolean value.

    - True: Use colors provided by Apprise
    - False: Do not use colors at all
    - rrggbb: where you provide the rgb values (hence #333333)
    - rgb: is also accepted as rgb values (hence #333)

    For RGB colors, the hashtag is optional
    """

    __color_rgb = re.compile(
        r"#?((?P<r1>[0-9A-F]{2})(?P<g1>[0-9A-F]{2})(?P<b1>[0-9A-F]{2})"
        r"|(?P<r2>[0-9A-F])(?P<g2>[0-9A-F])(?P<b2>[0-9A-F]))",
        re.IGNORECASE,
    )

    def __init__(self, color, asset=None):
        """Parses the color object accordingly."""

        # Initialize an asset object if one isn't otherwise defined
        self.asset = (
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()
        )

        # Prepare our color
        self.color = color
        if isinstance(color, str):
            self.color = self.__color_rgb.match(color)
            if self.color:
                # Store our RGB value as #rrggbb
                self.color = (
                    "{red}{green}{blue}".format(
                        red=self.color.group("r1"),
                        green=self.color.group("g1"),
                        blue=self.color.group("b1"),
                    ).lower()
                    if self.color.group("r1")
                    else "{red1}{red2}{green1}{green2}{blue1}{blue2}".format(
                        red1=self.color.group("r2"),
                        red2=self.color.group("r2"),
                        green1=self.color.group("g2"),
                        green2=self.color.group("g2"),
                        blue1=self.color.group("b2"),
                        blue2=self.color.group("b2"),
                    ).lower()
                )

        if self.color is None:
            # Color not determined, so base it on boolean parser
            self.color = parse_bool(color)

    def get(self, notify_type=NotifyType.INFO):
        """Returns color or true/false value based on configuration."""

        if isinstance(self.color, bool) and self.color:
            # We want to use the asset value
            return self.asset.color(notify_type=notify_type)

        elif self.color:
            # return our color as is
            return "#" + self.color

        # No color to return
        return None

    def __str__(self):
        """Our color representation."""
        if isinstance(self.color, bool):
            return "yes" if self.color else "no"

        # otherwise return our color
        return self.color

    def __bool__(self):
        """Allows this object to be wrapped in an 'if statement'.

        True is returned if a color was loaded
        """
        return bool(self.color is True or isinstance(self.color, str))



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a BulkSMS account
# You will need credits (new accounts start with a few)
#     https://www.bulksms.com/account/
#
# API is documented here:
#   - https://www.bulksms.com/developer/json/v1/#tag/Message
from itertools import chain
import json
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_bool, parse_phone_no
from .base import NotifyBase

IS_GROUP_RE = re.compile(
    r"^(@?(?P<group>[A-Z0-9_-]+))$",
    re.IGNORECASE,
)


class BulkSMSRoutingGroup:
    """The different categories of routing."""

    ECONOMY = "ECONOMY"
    STANDARD = "STANDARD"
    PREMIUM = "PREMIUM"


# Used for verification purposes
BULKSMS_ROUTING_GROUPS = (
    BulkSMSRoutingGroup.ECONOMY,
    BulkSMSRoutingGroup.STANDARD,
    BulkSMSRoutingGroup.PREMIUM,
)


class BulkSMSEncoding:
    """The different categories of routing."""

    TEXT = "TEXT"
    UNICODE = "UNICODE"
    BINARY = "BINARY"


class NotifyBulkSMS(NotifyBase):
    """A wrapper for BulkSMS Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "BulkSMS"

    # The services URL
    service_url = "https://bulksms.com/"

    # All notification requests are secure
    secure_protocol = "bulksms"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_bulksms"

    # BulkSMS uses the http protocol with JSON requests
    notify_url = "https://api.bulksms.com/v1/messages"

    # The maximum length of the body
    body_maxlen = 160

    # The maximum amount of texts that can go out in one batch
    default_batch_size = 4000

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{user}:{password}@{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "target_group": {
                "name": _("Target Group"),
                "type": "string",
                "prefix": "@",
                "regex": (r"^[A-Z0-9 _-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
            "route": {
                "name": _("Route Group"),
                "type": "choice:string",
                "values": BULKSMS_ROUTING_GROUPS,
                "default": BulkSMSRoutingGroup.STANDARD,
            },
            "unicode": {
                # Unicode characters
                "name": _("Unicode Characters"),
                "type": "bool",
                "default": True,
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self,
        source=None,
        targets=None,
        unicode=None,
        batch=None,
        route=None,
        **kwargs,
    ):
        """Initialize BulkSMS Object."""
        super().__init__(**kwargs)

        self.source = None
        if source:
            result = is_phone_no(source)
            if not result:
                msg = (
                    "The Account (From) Phone # specified "
                    f"({source}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            # Tidy source
            self.source = "+{}".format(result["full"])

        # Setup our route
        self.route = (
            self.template_args["route"]["default"]
            if not isinstance(route, str)
            else route.upper()
        )
        if self.route not in BULKSMS_ROUTING_GROUPS:
            msg = f"The route specified ({route}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Define whether or not we should set the unicode flag
        self.unicode = (
            self.template_args["unicode"]["default"]
            if unicode is None
            else bool(unicode)
        )

        # Define whether or not we should operate in a batch mode
        self.batch = (
            self.template_args["batch"]["default"]
            if batch is None
            else bool(batch)
        )

        # Parse our targets
        self.targets = []
        self.groups = []

        for target in parse_phone_no(targets):
            # Parse each phone number we found
            result = is_phone_no(target)
            if result:
                self.targets.append("+{}".format(result["full"]))
                continue

            group_re = IS_GROUP_RE.match(target)
            if group_re and not target.isdigit():
                # If the target specified is all digits, it MUST have a @
                # in front of it to eliminate any ambiguity
                self.groups.append(group_re.group("group"))
                continue

            self.logger.warning(
                f"Dropped invalid phone # and/or Group ({target}) specified.",
            )

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform BulkSMS Notification."""

        if not (self.password and self.user):
            self.logger.warning(
                "There were no valid login credentials provided"
            )
            return False

        if not (self.targets or self.groups):
            # We have nothing to notify
            self.logger.warning("There are no BulkSMS targets to notify")
            return False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Prepare our payload
        payload = {
            # The To gets populated in the loop below
            "to": None,
            "body": body,
            "routingGroup": self.route,
            "encoding": (
                BulkSMSEncoding.UNICODE
                if self.unicode
                else BulkSMSEncoding.TEXT
            ),
            # Options are NONE, ALL and ERRORS
            "deliveryReports": "ERRORS",
        }

        if self.source:
            payload.update({
                "from": self.source,
            })

        # Authentication
        auth = (self.user, self.password)

        # Prepare our targets
        targets = (
            list(self.targets)
            if batch_size == 1
            else [
                self.targets[index : index + batch_size]
                for index in range(0, len(self.targets), batch_size)
            ]
        )
        targets += [{"type": "GROUP", "name": g} for g in self.groups]

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload["to"] = target

            # Printable reference
            if isinstance(target, dict):
                p_target = target["name"]

            elif isinstance(target, list):
                p_target = f"{len(target)} targets"

            else:
                p_target = target

            # Some Debug Logging
            self.logger.debug(
                "BulkSMS POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"BulkSMS Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=json.dumps(payload),
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # The responsne might look like:
                # [
                #   {
                #       "id": "string",
                #       "type": "SENT",
                #       "from": "string",
                #       "to": "string",
                #       "body": null,
                #       "encoding": "TEXT",
                #       "protocolId": 0,
                #       "messageClass": 0,
                #       "numberOfParts": 0,
                #       "creditCost": 0,
                #       "submission": {...},
                #       "status": {...},
                #       "relatedSentMessageId": "string",
                #       "userSuppliedId": "string"
                #   }
                # ]

                if r.status_code not in (
                    requests.codes.created,
                    requests.codes.ok,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code
                    )

                    # set up our status code to use
                    status_code = r.status_code

                    self.logger.warning(
                        "Failed to send BulkSMS notification to {}: "
                        "{}{}error={}.".format(
                            p_target,
                            status_str,
                            ", " if status_str else "",
                            status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent BulkSMS notification to {p_target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending BulkSMS: to %s ",
                    p_target,
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "unicode": "yes" if self.unicode else "no",
            "batch": "yes" if self.batch else "no",
            "route": self.route,
        }

        if self.source:
            params["from"] = self.source

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{user}:{password}@{targets}/?{params}".format(
            schema=self.secure_protocol,
            user=self.pprint(self.user, privacy, safe=""),
            password=self.pprint(
                self.password, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            targets="/".join(
                chain(
                    [
                        NotifyBulkSMS.quote(f"{x}", safe="+")
                        for x in self.targets
                    ],
                    [
                        NotifyBulkSMS.quote(f"@{x}", safe="@")
                        for x in self.groups
                    ],
                )
            ),
            params=NotifyBulkSMS.urlencode(params),
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.user if self.user else None,
            self.password if self.password else None,
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""

        #
        # Factor batch into calculation
        #
        # Note: Groups always require a separate request (and can not be
        # included in batch calculations)
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets + len(self.groups)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = [
            NotifyBulkSMS.unquote(results["host"]),
            *NotifyBulkSMS.split_path(results["fullpath"]),
        ]

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyBulkSMS.unquote(results["qsd"]["from"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyBulkSMS.parse_phone_no(
                results["qsd"]["to"]
            )

        # Unicode Characters
        results["unicode"] = parse_bool(
            results["qsd"].get(
                "unicode", NotifyBulkSMS.template_args["unicode"]["default"]
            )
        )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyBulkSMS.template_args["batch"]["default"]
            )
        )

        # Allow one to define a route group
        if "route" in results["qsd"] and len(results["qsd"]["route"]):
            results["route"] = NotifyBulkSMS.unquote(results["qsd"]["route"])

        return results
